@@ -3,7 +3,8 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2012-2016, VU University Amsterdam
+    Copyright (c)  2012-2020, VU University Amsterdam
+                              CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -36,18 +37,25 @@
           [ prolog_walk_code/1,         % +Options
             prolog_program_clause/2     % -ClauseRef, +Options
           ]).
-:- use_module(library(option)).
-:- use_module(library(record)).
-:- use_module(library(debug)).
-:- use_module(library(apply)).
-:- use_module(library(lists)).
-:- use_module(library(prolog_metainference)).
+:- use_module(library(record),[(record)/1, op(_,_,record)]).
+
+:- autoload(library(apply),[maplist/2]).
+:- autoload(library(debug),[debug/3,debugging/1,assertion/1]).
+:- autoload(library(error),[must_be/2]).
+:- autoload(library(listing),[portray_clause/1]).
+:- autoload(library(lists),[member/2,nth1/3,append/3]).
+:- autoload(library(option),[meta_options/3]).
+:- autoload(library(prolog_clause),
+	    [clause_info/4,initialization_layout/4,clause_name/2]).
+:- autoload(library(prolog_metainference),
+	    [inferred_meta_predicate/2,infer_meta_predicate/2]).
+
 
 /** <module> Prolog code walker
 
 This module walks over  the  loaded   program,  searching  for  callable
 predicates. It started as part of  library(prolog_autoload) and has been
-turned into a seperate module to  facilitate operations that require the
+turned into a separate module to  facilitate operations that require the
 same reachability analysis, such as finding   references to a predicate,
 finding unreachable code, etc.
 
@@ -91,6 +99,7 @@ source file is passed into _Where.
                                                 test,development]))),
                        source(boolean),
                        trace_reference(any),
+                       trace_condition(callable),
                        on_trace(callable),
                        infer_meta_predicates(oneof([false,true,all])),
                        evaluate(boolean),
@@ -107,6 +116,7 @@ source file is passed into _Where.
                 infer_meta_predicates:oneof([false,true,all])=true,
                 clauses:list,               % Walk only these clauses
                 trace_reference:any=(-),
+                trace_condition:callable,   % Call-back condition
                 on_trace:callable,          % Call-back on trace hits
                                             % private stuff
                 clause,                     % Processed clause
@@ -150,8 +160,8 @@ source file is passed into _Where.
 %     * module(+Module)
 %     Only process the given module
 %
-%     * module_class(+ModuleClass)
-%     Limit processing to modules of this class. See
+%     * module_class(+ModuleClassList)
+%     Limit processing to modules of the given classes. See
 %     module_property/2 for details on module classes.  Default
 %     is to scan the classes =user= and =library=.
 %
@@ -168,6 +178,23 @@ source file is passed into _Where.
 %     represented as Module:Callable (i.e., they are always
 %     qualified).  See also subsumes_term/2.
 %
+%     * trace_condition(:Cond)
+%     Additional filter condition applied after `trace_reference`.
+%     Called as call(Cond, Callee, Context), where `Context` is a
+%     dict containing the following keys:
+%
+%       - Context:caller
+%         Qualified term representing the caller or the atom
+%         '<initialization>'.
+%       - Context:module
+%         Module being processed
+%       - Context:clause
+%         If we are processing a normal clause, the clause reference
+%         to this clause.
+%       - Context:initialization
+%         If we are processing an initialization/1 directive, a term
+%         `File:Line` representing the location of the declaration.
+%
 %     * on_trace(:OnTrace)
 %     If a reference to =trace_reference= is found, call
 %     call(OnTrace, Callee, Caller, Location), where Location is one
@@ -227,7 +254,7 @@ prolog_walk_code(Iteration, Options) :-
     ).
 
 is_meta(on_trace).
-
+is_meta(trace_condition).
 
 %!  walk_clauses(+Clauses, +OTerm) is det.
 %
@@ -251,10 +278,16 @@ walk_clauses(Clauses, OTerm) :-
 %
 %   True if we must scan Module according to OTerm.
 
+scan_module(M, OTerm) :-
+    walk_option_module(OTerm, M1),
+    nonvar(M1),
+    !,
+    \+ M \= M1.
 scan_module(M, OTerm) :-
     walk_option_module_class(OTerm, Classes),
     module_property(M, class(Class)),
-    memberchk(Class, Classes).
+    memberchk(Class, Classes),
+    !.
 
 %!  walk_from_initialization(+OTerm)
 %
@@ -265,10 +298,23 @@ scan_module(M, OTerm) :-
 
 walk_from_initialization(OTerm) :-
     walk_option_caller(OTerm, '<initialization>'),
-    forall('$init_goal'(_File, Goal, SourceLocation),
+    forall(init_goal_in_scope(Goal, SourceLocation, OTerm),
            ( walk_option_initialization(OTerm, SourceLocation),
              walk_from_initialization(Goal, OTerm))).
 
+init_goal_in_scope(Goal, SourceLocation, OTerm) :-
+    '$init_goal'(File, Goal, SourceLocation),
+    (   walk_option_module(OTerm, M),
+        nonvar(M)
+    ->  module_property(M, file(File))
+    ;   walk_option_module_class(OTerm, Classes),
+        source_file_property(File, module(MF))
+    ->  module_property(MF, class(Class)),
+        memberchk(Class, Classes),
+        walk_option_module(OTerm, MF)
+    ;   true
+    ).
+
 walk_from_initialization(M:Goal, OTerm) :-
     scan_module(M, OTerm),
     !,
@@ -283,6 +329,7 @@ walk_from_initialization(_, _).
 
 find_walk_from_module(M, OTerm) :-
     debug(autoload, 'Analysing module ~q', [M]),
+    walk_option_module(OTerm, M),
     forall(predicate_in_module(M, PI),
            walk_called_by_pred(M:PI, OTerm)).
 
@@ -486,6 +533,7 @@ walk_called(Goal, Module, TermPos, OTerm) :-
     ;   predicate_property(Module:Goal, imported_from(M2)),
         subsumes_term(To, M2:Goal)
     ),
+    trace_condition(M2:Goal, TermPos, OTerm),
     print_reference(M2:Goal, TermPos, trace, OTerm),
     fail.                                   % Continue search
 walk_called(Goal, Module, _, OTerm) :-
@@ -515,6 +563,20 @@ walk_called(Meta, M, term_position(_,E,_,_,ArgPosList), OTerm) :-
     walk_option_clause(OTerm, ClauseRef),
     register_possible_meta_clause(ClauseRef),
     walk_meta_call(1, Head, Meta, M, ArgPosList, E-E, OTerm).
+walk_called(Closure, _, _, _) :-
+    blob(Closure, closure),
+    !,
+    '$closure_predicate'(Closure, Module:Name/Arity),
+    functor(Head, Name, Arity),
+    '$get_predicate_attribute'(Module:Head, defined, 1).
+walk_called(ClosureCall, _, _, _) :-
+    compound(ClosureCall),
+    functor(ClosureCall, Closure, _),
+    blob(Closure, closure),
+    !,
+    '$closure_predicate'(Closure, Module:Name/Arity),
+    functor(Head, Name, Arity),
+    '$get_predicate_attribute'(Module:Head, defined, 1).
 walk_called(Goal, Module, _, _) :-
     nonvar(Module),
     '$get_predicate_attribute'(Module:Goal, defined, 1),
@@ -526,6 +588,29 @@ walk_called(Goal, Module, TermPos, OTerm) :-
 walk_called(Goal, _Module, TermPos, OTerm) :-
     not_callable(Goal, TermPos, OTerm).
 
+%!  trace_condition(:Callee, +TermPos, +OTerm) is semidet.
+%
+%   Call call(Condition, Callee, Dict)
+
+trace_condition(Callee, TermPos, OTerm) :-
+    walk_option_trace_condition(OTerm, Cond), nonvar(Cond),
+    !,
+    cond_location_context(OTerm, TermPos, Context0),
+    walk_option_caller(OTerm, Caller),
+    walk_option_module(OTerm, Module),
+    put_dict(#{caller:Caller, module:Module}, Context0, Context),
+    call(Cond, Callee, Context).
+trace_condition(_, _, _).
+
+cond_location_context(OTerm, _TermPos, Context) :-
+    walk_option_clause(OTerm, Clause), nonvar(Clause),
+    !,
+    Context = #{clause:Clause}.
+cond_location_context(OTerm, _TermPos, Context) :-
+    walk_option_initialization(OTerm, Init), nonvar(Init),
+    !,
+    Context = #{initialization:Init}.
+
 %!  undecided(+Variable, +TermPos, +OTerm)
 
 undecided(Var, TermPos, OTerm) :-
@@ -852,6 +937,8 @@ walk_dcg_body(G, M, TermPos, OTerm) :-
 :- meta_predicate
     subterm_pos(+, +, 2, +, -),
     sublist_pos(+, +, +, +, 2, -).
+:- public
+    subterm_pos/5.                      % used in library(check).
 
 subterm_pos(_, _, _, Pos, _) :-
     var(Pos), !, fail.
@@ -903,6 +990,17 @@ extend(Goal, N, GoalEx, TermPosIn, TermPosOut, _) :-
     extend_term_pos(TermPosIn, N, TermPosOut),
     append(List, Extra, ListEx),
     GoalEx =.. ListEx.
+extend(Closure, N, M:GoalEx, TermPosIn, TermPosOut, OTerm) :-
+    blob(Closure, closure),             % call(Closure, A1, ...)
+    !,
+    '$closure_predicate'(Closure, M:Name/Arity),
+    length(Extra, N),
+    extend_term_pos(TermPosIn, N, TermPosOut),
+    GoalEx =.. [Name|Extra],
+    (   N =:= Arity
+    ->  true
+    ;   print_reference(Closure, TermPosIn, closure_arity_mismatch, OTerm)
+    ).
 extend(Goal, _, _, TermPos, _, OTerm) :-
     print_reference(Goal, TermPos, not_callable, OTerm).
 
@@ -942,10 +1040,15 @@ variants([H|T], V, List) :-
 predicate_in_module(Module, PI) :-
     current_predicate(Module:PI),
     PI = Name/Arity,
+    \+ hidden_predicate(Name, Arity),
     functor(Head, Name, Arity),
     \+ predicate_property(Module:Head, imported_from(_)).
 
 
+hidden_predicate(Name, _) :-
+    sub_atom(Name, 0, _, _, '$wrap$').
+
+
                  /*******************************
                  *      ENUMERATE CLAUSES       *
                  *******************************/
@@ -1019,7 +1122,7 @@ initialization_clause(ClauseRef, OTerm) :-
 
 prolog:message(trace_call_to(PI, Context)) -->
     [ 'Call to ~q at '-[PI] ],
-    prolog:message_location(Context).
+    '$messages':swi_location(Context).
 
 prolog:message_location(clause_term_position(ClauseRef, TermPos)) -->
     { clause_property(ClauseRef, file(File)) },
